import 'dart:math';

import 'package:sudoku/assets/grids.dart';
import 'package:sudoku/config/default_global_settings.dart';
import 'package:sudoku/cubit/game_cubit.dart';
import 'package:sudoku/models/board.dart';
import 'package:sudoku/models/cell.dart';
import 'package:sudoku/models/cell_location.dart';
import 'package:sudoku/models/settings_game.dart';
import 'package:sudoku/models/settings_global.dart';
import 'package:sudoku/models/types.dart';
import 'package:sudoku/utils/board_utils.dart';
import 'package:sudoku/utils/tools.dart';

class Game {
  Game({
    required this.gameSettings,
    required this.globalSettings,
    required this.board,
    required this.solvedBoard,
    required this.isRunning,
    required this.isFinished,
    this.shuffledCellValues = const [],
    this.boardConflicts = const [],
    this.selectedCell,
    this.showConflicts = false,
    this.givenTipsCount = 0,
    this.buttonTipsCountdown = 0,
    this.animationInProgress = false,
    this.boardAnimated = const [],
    required this.blockSizeHorizontal,
    required this.blockSizeVertical,
    required this.boardSize,
  });

  final GameSettings gameSettings;
  final GlobalSettings globalSettings;

  bool isRunning = false;
  bool isFinished = false;

  int blockSizeVertical = 0;
  int blockSizeHorizontal = 0;
  int boardSize = 0;

  Board board;
  Board solvedBoard;

  List<int> shuffledCellValues = [];
  Cell? selectedCell;

  int givenTipsCount = 0;
  int buttonTipsCountdown = 0;
  bool showConflicts = false;
  ConflictsCount boardConflicts = [];

  bool animationInProgress = false;
  AnimatedBoard boardAnimated = [];

  factory Game.createNull() {
    return Game(
      gameSettings: GameSettings.createDefault(),
      globalSettings: GlobalSettings.createDefault(),
      board: Board.createEmpty(),
      solvedBoard: Board.createEmpty(),
      shuffledCellValues: [],
      isRunning: false,
      isFinished: false,
      givenTipsCount: 0,
      blockSizeHorizontal: 0,
      blockSizeVertical: 0,
      boardSize: 0,
    );
  }

  factory Game.createNew({
    GameSettings? gameSettings,
    GlobalSettings? globalSettings,
  }) {
    final GameSettings newGameSettings = gameSettings ?? GameSettings.createDefault();
    final GlobalSettings newGlobalSettings = globalSettings ?? GlobalSettings.createDefault();

    final int blockSizeHorizontal = int.parse(newGameSettings.size.split('x')[0]);
    final int blockSizeVertical = int.parse(newGameSettings.size.split('x')[1]);
    final int boardSize = blockSizeHorizontal * blockSizeVertical;

    const int maxCellValue = 16;
    final List<int> shuffledCellValues = List<int>.generate(maxCellValue, (i) => i + 1);

    if (DefaultGlobalSettings.shufflableSkins.contains(newGlobalSettings.skin)) {
      shuffledCellValues.shuffle();
      printlog('Shuffled tiles values: $shuffledCellValues');
    }

    ConflictsCount nonConflictedBoard = [];

    for (int row = 0; row < boardSize; row++) {
      List<int> line = [];
      for (int col = 0; col < boardSize; col++) {
        line.add(0);
      }
      nonConflictedBoard.add(line);
    }

    final List<String> templates =
        SudokuGrids.templates[newGameSettings.size]?[newGameSettings.level] ?? [];
    final String template = templates.elementAt(Random().nextInt(templates.length)).toString();

    if (template.length != pow(blockSizeHorizontal * blockSizeVertical, 2)) {
      printlog('Failed to get grid template...');
      return Game.createNull();
    }

    final Board board = BoardUtils.createBoardFromTemplate(
      template: template,
      isSymetric: (blockSizeHorizontal == blockSizeVertical),
    );
    final Board solvedBoard = Board.createNew(cells: board.getSolvedGrid());

    // Animated background
    AnimatedBoard notAnimatedBoard = [];
    for (int row = 0; row < boardSize; row++) {
      List<bool> line = [];
      for (int col = 0; col < boardSize; col++) {
        line.add(false);
      }
      notAnimatedBoard.add(line);
    }

    return Game(
      gameSettings: newGameSettings,
      globalSettings: newGlobalSettings,
      board: board,
      solvedBoard: solvedBoard,
      isRunning: true,
      isFinished: false,
      boardConflicts: nonConflictedBoard,
      shuffledCellValues: shuffledCellValues,
      selectedCell: null,
      blockSizeHorizontal: blockSizeHorizontal,
      blockSizeVertical: blockSizeVertical,
      boardSize: boardSize,
      boardAnimated: notAnimatedBoard,
    );
  }

  bool canGiveTip() {
    return (buttonTipsCountdown == 0);
  }

  int getTranslatedValueForDisplay(int originalValue) {
    return shuffledCellValues[originalValue - 1];
  }

  bool checkBoardIsSolved() {
    // (re)compute conflicts
    boardConflicts = computeConflictsInBoard();

    // check grid is fully completed and does not contain conflict
    for (int row = 0; row < boardSize; row++) {
      for (int col = 0; col < boardSize; col++) {
        if (board.cells[row][col].value == 0 || boardConflicts[row][col] != 0) {
          return false;
        }
      }
    }

    printlog('-> ok sudoku solved!');

    return true;
  }

  ConflictsCount computeConflictsInBoard() {
    final BoardCells cells = board.cells;
    final ConflictsCount conflicts = boardConflicts;

    // reset conflict states
    for (int row = 0; row < boardSize; row++) {
      for (int col = 0; col < boardSize; col++) {
        conflicts[row][col] = 0;
      }
    }

    // check lines does not contains a value twice
    for (int row = 0; row < boardSize; row++) {
      final List<int> values = [];
      for (int col = 0; col < boardSize; col++) {
        int value = cells[row][col].value;
        if (value != 0) {
          values.add(value);
        }
      }
      final List<int> distinctValues = values.toSet().toList();
      if (values.length != distinctValues.length) {
        printlog('line $row contains duplicates');
        // Add line to cells in conflict
        for (int col = 0; col < boardSize; col++) {
          conflicts[row][col]++;
        }
      }
    }

    // check columns does not contains a value twice
    for (int col = 0; col < boardSize; col++) {
      final List<int> values = [];
      for (int row = 0; row < boardSize; row++) {
        int value = cells[row][col].value;
        if (value != 0) {
          values.add(value);
        }
      }
      final List<int> distinctValues = values.toSet().toList();
      if (values.length != distinctValues.length) {
        printlog('column $col contains duplicates');
        // Add column to cells in conflict
        for (int row = 0; row < boardSize; row++) {
          conflicts[row][col]++;
        }
      }
    }

    // check blocks does not contains a value twice
    final int horizontalBlocksCount = blockSizeVertical;
    final int verticalBlocksCount = blockSizeHorizontal;
    for (int blockRow = 0; blockRow < verticalBlocksCount; blockRow++) {
      for (int blockCol = 0; blockCol < horizontalBlocksCount; blockCol++) {
        List<int> values = [];

        for (int rowInBlock = 0; rowInBlock < blockSizeVertical; rowInBlock++) {
          for (int colInBlock = 0; colInBlock < blockSizeHorizontal; colInBlock++) {
            int row = (blockRow * blockSizeVertical) + rowInBlock;
            int col = (blockCol * blockSizeHorizontal) + colInBlock;
            int value = cells[row][col].value;
            if (value != 0) {
              values.add(value);
            }
          }
        }

        List<int> distinctValues = values.toSet().toList();
        if (values.length != distinctValues.length) {
          printlog('block [$blockCol,$blockRow] contains duplicates');
          // Add blocks to cells in conflict
          for (int rowInBlock = 0; rowInBlock < blockSizeVertical; rowInBlock++) {
            for (int colInBlock = 0; colInBlock < blockSizeHorizontal; colInBlock++) {
              int row = (blockRow * blockSizeVertical) + rowInBlock;
              int col = (blockCol * blockSizeHorizontal) + colInBlock;
              conflicts[row][col]++;
            }
          }
        }
      }
    }

    return conflicts;
  }

  void showTip(GameCubit gameCubit) {
    if (selectedCell == null) {
      // no selected cell -> pick one
      helpSelectCell(gameCubit);
    } else {
      // currently selected cell -> set value
      helpFillCell(gameCubit);
    }
    gameCubit.increaseGivenTipsCount();
  }

  void helpSelectCell(GameCubit gameCubit) {
    // pick one of wrong value cells, if found
    final List<List<int>> wrongValueCells = getCellsWithWrongValue();
    if (wrongValueCells.isNotEmpty) {
      printlog('will pick from wrongValueCells');
      pickRandomFromList(gameCubit, wrongValueCells);
      return;
    }

    // pick one of conflicting cells, if found
    final List<List<int>> conflictingCells = getCellsWithConflicts();
    if (conflictingCells.isNotEmpty) {
      printlog('will pick from conflictingCells');
      pickRandomFromList(gameCubit, conflictingCells);
      return;
    }

    //  pick one form cells with unique non-conflicting candidate value
    final List<List<int>> candidateCells = board.getEmptyCellsWithUniqueAvailableValue();
    if (candidateCells.isNotEmpty) {
      printlog('will pick from candidateCells');
      pickRandomFromList(gameCubit, candidateCells);
      return;
    }
  }

  List<List<int>> getCellsWithWrongValue() {
    final BoardCells cells = board.cells;
    final BoardCells cellsSolved = solvedBoard.cells;

    List<List<int>> cellsWithWrongValue = [];

    for (int row = 0; row < boardSize; row++) {
      for (int col = 0; col < boardSize; col++) {
        if (cells[row][col].value != 0 &&
            cells[row][col].value != cellsSolved[row][col].value) {
          cellsWithWrongValue.add([row, col]);
        }
      }
    }

    return cellsWithWrongValue;
  }

  List<List<int>> getCellsWithConflicts() {
    List<List<int>> cellsWithConflict = [];

    for (int row = 0; row < boardSize; row++) {
      for (int col = 0; col < boardSize; col++) {
        if (boardConflicts[row][col] != 0) {
          cellsWithConflict.add([row, col]);
        }
      }
    }

    return cellsWithConflict;
  }

  void pickRandomFromList(GameCubit gameCubit, List<List<int>> cellsCoordinates) {
    if (cellsCoordinates.isNotEmpty) {
      cellsCoordinates.shuffle();
      final List<int> cell = cellsCoordinates[0];
      gameCubit.selectCell(CellLocation.go(cell[0], cell[1]));
    }
  }

  void helpFillCell(GameCubit gameCubit) {
    // Will clean cell if no eligible value found
    int eligibleValue = 0;

    // Ensure there is only one eligible value for this cell
    int allowedValuesCount = 0;
    for (int value = 1; value <= boardSize; value++) {
      if (board.isValueAllowed(selectedCell?.location, value)) {
        allowedValuesCount++;
        eligibleValue = value;
      }
    }

    gameCubit.updateCellValue(
        selectedCell!.location, allowedValuesCount == 1 ? eligibleValue : 0);
    gameCubit.unselectCell();
  }

  printGrid() {
    final BoardCells cells = board.cells;
    final BoardCells solvedCells = solvedBoard.cells;

    const String stringValues = '0123456789ABCDEFG';
    printlog('');
    printlog('-------');
    for (int rowIndex = 0; rowIndex < cells.length; rowIndex++) {
      String row = '';
      String rowSolved = '';
      for (int colIndex = 0; colIndex < cells[rowIndex].length; colIndex++) {
        row += stringValues[cells[rowIndex][colIndex].value];
        rowSolved += stringValues[solvedCells[rowIndex][colIndex].value];
      }
      printlog('$row | $rowSolved');
    }
    printlog('-------');
    printlog('');
  }

  void dump() {
    printlog('');
    printlog('## Current game dump:');
    printlog('');
    gameSettings.dump();
    globalSettings.dump();
    printlog('');
    printlog('$Game:');
    printlog('  blockSizeHorizontal: $blockSizeHorizontal');
    printlog('  blockSizeVertical: $blockSizeVertical');
    printlog('  boardSize: $boardSize');
    printlog('  shuffledCellValues: $shuffledCellValues');
    printlog('');
    printlog('  isRunning: $isRunning');
    printlog('  isFinished: $isFinished');
    printlog('  selectedCell: ${selectedCell?.toString() ?? ''}');
    printlog('  showConflicts: $showConflicts');
    printlog('  givenTipsCount: $givenTipsCount');
    printlog('  givenTipsCountEnableCountdown: $buttonTipsCountdown');
    printlog('  animationInProgress: $animationInProgress');
    printlog('');

    printGrid();
    printlog('');
  }

  @override
  String toString() {
    return '$Game(${toJson()})';
  }

  Map<String, dynamic>? toJson() {
    return <String, dynamic>{
      'gameSettings': gameSettings.toJson(),
      'globalSettings': globalSettings.toJson(),
      'isRunning': isRunning,
      'isFinished': isFinished,
      'board': board.toJson(),
      'solvedBoard': solvedBoard.toJson(),
      'shuffledCellValues': shuffledCellValues,
      'boardConflicts': boardConflicts,
      'selectedCell': selectedCell?.toJson(),
      'showConflicts': showConflicts,
      'givenTipsCount': givenTipsCount,
      'givenTipsCountEnableCountdown': buttonTipsCountdown,
      'animationInProgress': animationInProgress,
      'boardAnimated': boardAnimated,
      'blockSizeHorizontal': blockSizeHorizontal,
      'blockSizeVertical': blockSizeVertical,
    };
  }
}
